Attribute VB_Name = "GraFX_001_GRAPHIC"
Option Explicit

'===API===

'Mauszeiger an / aus
Private Declare Function ShowCursor Lib "user32" (ByVal bShow As Long) As Long

'Screenshot
Private Declare Function GetDC Lib "user32" (ByVal hWnd As Long) As Long
Private Declare Function ReleaseDC Lib "user32" (ByVal hWnd As Long, ByVal hdc As Long) As Long
Private Declare Function GetDesktopWindow Lib "user32" () As Long

'zum Blitten des Screenshots auf PictureBox
Private Declare Function BitBlt Lib "gdi32" (ByVal hDestDC As Long, ByVal X As Long, _
ByVal Y As Long, ByVal nWidth As Long, ByVal nHeight As Long, ByVal hSrcDC As Long, _
ByVal xSrc As Long, ByVal ySrc As Long, ByVal dwRop As Long) As Long

'fr Time-Counter
Private Declare Function QueryPerformanceFrequency Lib "kernel32" _
(ByRef Frequency As Currency) As Long

Private Declare Function QueryPerformanceCounter Lib "kernel32" _
(ByRef Counter As Currency) As Long


'===Enums===

Public Enum ETextAlignment                                  'Block-Text-Ausrichtung
    TEXT_ALIGNMENT_LEFT = 1
    TEXT_ALIGNMENT_RIGHT
    TEXT_ALIGNMENT_CENTERED
End Enum

Public Enum EDisplayMode                                    'Display-Modus
    DISP_MODE_FULLSCREEN = 1
    DISP_MODE_WINDOW
End Enum

Public Enum EColorTransform
    ONLY_START_COLOR = 1
    START_AND_INV_COLOR
    ALL_COLORS
End Enum

Private Enum EAlphaEffect
    A_OFF = 1
    A_ADD
    A_MULTIPLY
    A_SUBTRACT
    A_INVERT
End Enum


'===Typendefinitionen===

Public Type TSurf                                           'Surface
    Surf                                As DirectDrawSurface7
    Width                               As Integer
    Height                              As Integer
End Type

Public Type TAnim                                           'Animation
    NumFrames                           As Integer
    FrameWidth                          As Integer
    ActFrameSng                         As Single
    ActFrameInt                         As Integer
End Type

Public Type TFont                                           'Bitmap-Schrift
    Surf                                As TSurf
    Letter(32 To 255)                   As RECT
End Type

Private Type TMsg                                           'Blit-Positions-unabhngige Nachricht
    Text                                As String
    X                                   As Integer
    Y                                   As Integer
End Type

Public Type TRectSng                                        'Rectangle mit Single-Koordinaten
    Left                                As Single
    Top                                 As Single
    Right                               As Single
    Bottom                              As Single
End Type

Private Type TEffectParticle
    VX                                  As Single
    VY                                  As Single
    MX                                  As Single
    MY                                  As Single
    Rad                                 As Single
    RS                                  As Single
    Attenuation                         As Single
    Radius                              As Single
End Type

Public Type TEffect
    Draw                        As Boolean
    
    InfiniteSpread              As Boolean
    CurrentParticle             As Long
    CurAddParticle              As Single
    
    ParticleCount               As Long
   
    X                           As Single
    Y                           As Single
    VelX                        As Single
    VelY                        As Single
    AccX                        As Single
    AccY                        As Single
    
    StartMinRadius              As Single
    StartMaxRadius              As Single
    EndRadius                   As Single
    
    SpreadRangeOffset           As Single
    SpreadRange                 As Single
        
    VMaxIncreasing              As Single
    VMinIncreasing              As Single
    VDivIncreasing              As Boolean
    
    VDecelThreshold             As Single
    VDeceleration               As Single
    
    ColorMode                   As EColorTransform
    StartColor                  As D3DCOLORVALUE
    InvColor                    As D3DCOLORVALUE
    EndColor                    As D3DCOLORVALUE
    
    MinParticleRadius           As Single
    MaxParticleRadius           As Single
    
    MinParticleRot              As Single
    MaxParticleRot              As Single
    
    AttenuationStartRange       As Single
    AttenuationSpeed            As Single
    AttenuationExpMode          As Boolean

    Particle()                  As TEffectParticle
End Type

Public Type TLight
    Draw                        As Boolean
    X                           As Single
    Y                           As Single
    VelX                        As Single
    VelY                        As Single
    AccX                        As Single
    AccY                        As Single
    
    Radius                      As Single
    iRadius                     As Single
    
    Color                       As D3DCOLORVALUE
    Intensity                   As Single
    
    AttenuationSpeed            As Single
    Attenuation                 As Single
    
    VDecelThreshold             As Single
    VDeceleration               As Single
End Type

Private Type TWaveNetParticle
    Vertex(1 To 4)              As D3DVERTEX
    Height                      As Single
    Draw                        As Boolean
End Type

Public Type T3DWave
    Draw                        As Boolean

    Amplitude                   As Single
    AmpAttenuation              As Single
    WaveLength                  As Single
    WaveCount                   As Single
    MaxRad                      As Single
    Rad                         As Single
    
    c                           As Single
    cAttenuation                As Single
    
    X                           As Single
    Y                           As Single
    SrcX                        As Long
    SrcY                        As Long
End Type

Private Type TBlurNetParticle
    Vertex(1 To 4)              As D3DVERTEX
End Type


'===ffentliche Variablen===

Public g_DD                                 As DirectDraw7              'DirectDraw-Objekt
Public g_FrontBuf                           As DirectDrawSurface7       'FrontBuffer
Public g_BackBuf                            As DirectDrawSurface7       'BackBuffer
Public g_D3D                                As Direct3D7                'D3D-Objekt
Public g_D3DDev                             As Direct3DDevice7          'D3D-Device
Public g_RectViewport(0)                    As D3DRECT                  'Viewport
Public g_matProj                            As D3DMATRIX                'Projektionsmatrix
Public g_matView                            As D3DMATRIX                'ViewMatrix


'===interne Variablen===

Private iXRes                               As Integer
Private iYRes                               As Integer
Private iColDepth                           As Byte
Private iDispMode                           As EDisplayMode
Private ScreenRect                          As RECT
Private AVSpeed                             As Single
Private MsgQueue()                          As TMsg
Private MsgCount                            As Long
Private TimerFreq                           As Currency                 'Timer-Frequenz
Private LastCntrVal                         As Currency                 'letzte Zeitnahme
Private FPSLoopQuotient                     As Long
Private FramesPerSec                        As Single
Private Material                            As D3DMATERIAL7
Private DDI                                 As DirectDrawIdentifier
Private FXVertex(1 To 4)                    As D3DVERTEX
Private D3DDivX                             As Single
Private D3DDivY                             As Single
Private D3DSubX                             As Single
Private D3DSubY                             As Single
Private MAX_SNOW_PARTICLES                  As Long
Private SnowParticle()                      As TEffectParticle
Private SnowTexture                         As DirectDrawSurface7

'ShockWave
Private Const ELEMENT_WIDTH                 As Long = 16
Private Const ELEMENT_HEIGHT                As Long = 16
Private Const TEXTURE_WIDTH                 As Long = 256
Private Const TEXTURE_HEIGHT                As Long = 256
Private WaveNetV()                          As TWaveNetParticle
Private WaveNetT()                          As DirectDrawSurface7
Private HNET_ELEMENTS                       As Long
Private VNET_ELEMENTS                       As Long
Private HT_ELEMENTS                         As Long
Private VT_ELEMENTS                         As Long
Private ScreenRadius                        As Single

'MotionBlur
Private BlurNetV()                          As TBlurNetParticle
Private BlurNetT()                          As DirectDrawSurface7

Private Const PI                            As Single = 3.14159
Private Const HPI                           As Single = 1.57079

'initialisiert die Graphik-Engine
Public Function Init_GraFX_GRAPHIC(ByRef Frm As Form, Optional ByVal ResX As Integer = 1024, Optional ByVal ResY As Integer = 768, Optional ByVal ColDepth As Byte = 16, Optional ByVal DisplayMode As EDisplayMode = DISP_MODE_FULLSCREEN) As Boolean
        
    If DisplayMode = DISP_MODE_FULLSCREEN Then
        If Frm.WindowState > 0 Then
            MsgBox "Warning: WindowState needs to be 0 !!!"
            Exit Function
        End If
        
        If Frm.BorderStyle > 0 Then
            MsgBox "Warning: BorderStyle needs to be 0 !!!"
            Exit Function
        End If
    End If

    iDispMode = DisplayMode
    MsgCount = 0
    QueryPerformanceFrequency TimerFreq
    LastCntrVal = 0
    FPSLoopQuotient = 1
    FramesPerSec = 100
    
    Init_GraFX_GRAPHIC = Init_DD(Frm, ResX, ResY, ColDepth, DisplayMode)
    If DisplayMode = DISP_MODE_FULLSCREEN Then Init_GraFX_GRAPHIC = Init_GraFX_GRAPHIC And Init_D3D(ResX, ResY)

End Function

'fhrt die Graphik-Engine herunter
Public Sub Unload_GraFX_GRAPHIC(ByRef Frm As Form)

    Unload_DD Frm
    If iDispMode = DISP_MODE_FULLSCREEN Then Unload_D3D

End Sub

'Initialisiert DirectDraw und gibt zurck ob Vorgang erfolgreich war
Private Function Init_DD(ByRef Frm As Form, Optional ByVal ResX As Integer = 1024, Optional ByVal ResY As Integer = 768, Optional ByVal ColDepth As Byte = 16, Optional ByVal DisplayMode As EDisplayMode = DISP_MODE_FULLSCREEN) As Boolean

    Dim BufDesc             As DDSURFACEDESC2
    Dim Caps                As DDSCAPS2
    Dim h_ScreenSurf        As DDSURFACEDESC2
    Dim Clipper             As DirectDrawClipper
        
    On Error GoTo error:
        
    Set g_DD = g_DX.DirectDrawCreate("")
    
    'Focus auf Form setzen
    Frm.Show
    
    If DisplayMode = DISP_MODE_FULLSCREEN Then
        
        g_DD.SetCooperativeLevel Frm.hWnd, DDSCL_EXCLUSIVE Or DDSCL_FULLSCREEN
        g_DD.SetDisplayMode ResX, ResY, ColDepth, 0, DDSDM_DEFAULT
        
    ElseIf DisplayMode = DISP_MODE_WINDOW Then
    
        'Breite des Bildschirms ermitteln
        g_DD.GetDisplayMode h_ScreenSurf
        ResX = h_ScreenSurf.lWidth
        ResY = h_ScreenSurf.lHeight
        ColDepth = h_ScreenSurf.ddpfPixelFormat.lRGBBitCount
        
        g_DD.SetCooperativeLevel Frm.hWnd, DDSCL_NORMAL
        
    End If
            
    iXRes = ResX
    iYRes = ResY
    iColDepth = ColDepth
    ScreenRect = SetRect(0, 0, ResX, ResY)
            
    'Komplexe Surface erstellen
    With BufDesc
        .lFlags = DDSD_CAPS
        .ddsCaps.lCaps = DDSCAPS_PRIMARYSURFACE
        
        'nur fr Fullscreen
        If DisplayMode = DISP_MODE_FULLSCREEN Then
            .lFlags = .lFlags Or DDSD_BACKBUFFERCOUNT
            .ddsCaps.lCaps = .ddsCaps.lCaps Or DDSCAPS_3DDEVICE Or DDSCAPS_FLIP Or _
                             DDSCAPS_COMPLEX Or DDSCAPS_VIDEOMEMORY
            .lBackBufferCount = 1
        End If
        
        Set g_FrontBuf = g_DD.CreateSurface(BufDesc)
    End With
        
    'Backbuffer beschreiben (nur bei Fenstermodus)
    If DisplayMode = DISP_MODE_WINDOW Then
        
        With BufDesc
            .lFlags = DDSD_CAPS Or DDSD_HEIGHT Or DDSD_WIDTH
            .ddsCaps.lCaps = DDSCAPS_OFFSCREENPLAIN
            .lWidth = ResX
            .lHeight = ResY
        End With
        
        Set g_BackBuf = g_DD.CreateSurface(BufDesc)
        
        'Clipper initialisieren
        Set Clipper = g_DD.CreateClipper(0)
        
        'Clipper auf Form und Frontbuffer setzen
        Clipper.SetHWnd Frm.hWnd
        g_FrontBuf.SetClipper Clipper
        
    ElseIf DisplayMode = DISP_MODE_FULLSCREEN Then
    
        'BackBuffer mit Frontbuffer verbinden
        Caps.lCaps = DDSCAPS_BACKBUFFER
        Set g_BackBuf = g_FrontBuf.GetAttachedSurface(Caps)
    
    End If

    'Schrifteigenschaften setzen
    g_BackBuf.SetFontTransparency True
    g_BackBuf.SetForeColor vbWhite

    Init_DD = True
    Exit Function
    
error:
    Init_DD = False
    
End Function

'fhrt DirectDraw herunter
Private Sub Unload_DD(ByRef Frm As Form)

    'Auflsung wiederherstellen
    g_DD.RestoreDisplayMode
    g_DD.SetCooperativeLevel Frm.hWnd, DDSCL_NORMAL
    
    Set g_BackBuf = Nothing
    Set g_FrontBuf = Nothing
    Set g_DD = Nothing

End Sub

'ldt eine DirectDraw-Surface und gibt zurck ob Vorgang erfolgreich war
Public Function Load_Surf(ByVal Path As String, ByRef Surf As TSurf, Optional ByVal KeyCol As Long = -1, Optional ByVal Width As Integer = -1, Optional ByVal Height As Integer = -1) As Boolean

    Dim SurfDesc        As DDSURFACEDESC2
    Dim ColKey          As DDCOLORKEY
        
    On Error GoTo error:
    
    'Surface-Description
    With SurfDesc
        .lFlags = DDSD_CAPS
        
        If Width > 0 And Height > 0 Then
            .lFlags = .lFlags Or DDSD_WIDTH Or DDSD_HEIGHT
            .lWidth = Width
            .lHeight = Height
        End If
        
        .ddsCaps.lCaps = DDSCAPS_OFFSCREENPLAIN
    End With
    
    Set Surf.Surf = g_DD.CreateSurfaceFromFile(Path, SurfDesc)     'Bild laden
    
    Surf.Width = SurfDesc.lWidth
    Surf.Height = SurfDesc.lHeight
    
    'ColorKey setzen
    ColKey.low = KeyCol
    ColKey.high = KeyCol
    Surf.Surf.SetColorKey DDCKEY_SRCBLT, ColKey
    
    Load_Surf = True
    Exit Function
    
error:
    Load_Surf = False
    
End Function

'lscht eine Surface aus dem Speicher
Public Sub Unload_Surf(ByRef Surf As TSurf)

    Set Surf.Surf = Nothing

End Sub

'gibt ein durch Hhe und Breite definiertes Rectangle zurck
Public Function SetRectFrame(ByVal Left As Long, ByVal Top As Long, ByVal Width As Long, ByVal Height As Long) As RECT
    
    With SetRectFrame
        .Left = Left
        .Top = Top
        .Right = .Left + Width
        .Bottom = .Top + Height
    End With
    
End Function

'gibt ein Rectangle zurck
Public Function SetRect(ByVal Left As Integer, ByVal Top As Integer, ByVal Right As Integer, ByVal Bottom As Integer) As RECT
    
    With SetRect
        .Left = Left
        .Top = Top
        .Right = Right
        .Bottom = Bottom
    End With
    
End Function

'Zeichnet eine Surface
Public Sub Blit_Surf(ByVal X As Integer, ByVal Y As Integer, ByRef Surf As TSurf)

    Dim h_Rect              As RECT
    Dim h_Right             As Integer
    Dim h_Bottom            As Integer

    With h_Rect
    
        .Left = 0
        .Top = 0
        .Right = Surf.Width
        .Bottom = Surf.Height
            
        'rechter Rand
        h_Right = X + Surf.Width
        If h_Right > iXRes Then .Right = .Right - (h_Right - iXRes)
    
        'unterer Rand
        h_Bottom = Y + Surf.Height
        If h_Bottom > iYRes Then .Bottom = .Bottom - (h_Bottom - iYRes)
    
        'linker Rand
        If X < 0 Then
            .Left = .Left - X
            X = 0
        End If
    
        'oberer Rand
        If Y < 0 Then
            .Top = .Top - Y
            Y = 0
        End If
        
    End With
        
    'zeichnen
    g_BackBuf.BltFast X, Y, Surf.Surf, h_Rect, DDBLTFAST_DONOTWAIT Or DDBLTFAST_SRCCOLORKEY

End Sub

'Zeichnet eine Surface unter Angabe eines Rectangles
Public Sub Blit_SurfSector(ByVal X As Integer, ByVal Y As Integer, ByRef Surf As TSurf, ByRef Rectangle As RECT)

    Dim h_Rect              As RECT
    Dim h_Right             As Integer
    Dim h_Bottom            As Integer

    h_Rect = Rectangle
        
    With h_Rect
        
        'rechter Rand
        h_Right = X + (.Right - .Left)
        If h_Right > iXRes Then .Right = .Right - (h_Right - iXRes)
    
        'unterer Rand
        h_Bottom = Y + (.Bottom - .Top)
        If h_Bottom > iYRes Then .Bottom = .Bottom - (h_Bottom - iYRes)
    
        'linker Rand
        If X < 0 Then
            .Left = .Left - X
            X = 0
        End If
    
        'oberer Rand
        If Y < 0 Then
            .Top = .Top - Y
            Y = 0
        End If
        
    End With
    
    'zeichnen
    g_BackBuf.BltFast X, Y, Surf.Surf, h_Rect, DDBLTFAST_DONOTWAIT Or DDBLTFAST_SRCCOLORKEY

End Sub

'Initialisiert die Eigenschaften einer Animation
Public Sub Init_Anim(ByRef Surf As TSurf, ByRef Anim As TAnim, ByVal NumFrames As Integer)

    With Anim
        .NumFrames = NumFrames
        .FrameWidth = Surf.Width \ NumFrames
        .ActFrameSng = 1
        .ActFrameInt = 1
    End With

End Sub

'setzt ein AnimationsFrame
Public Sub SetAnimFrame(ByRef Anim As TAnim, ByVal Frame As Single)
    
    With Anim
        .ActFrameSng = Frame
        
        Do While .ActFrameSng < 1
            .ActFrameSng = .ActFrameSng + .NumFrames
        Loop
        
        Do While .ActFrameSng >= .NumFrames + 1
            .ActFrameSng = .ActFrameSng - .NumFrames
        Loop
                    
        .ActFrameInt = Int(.ActFrameSng)
    End With

End Sub

'animiert eine Animation
Public Sub Proceed_Anim(ByRef Anim As TAnim, ByVal Speed As Single)

    With Anim
        .ActFrameSng = .ActFrameSng + Speed
                            
        Do While .ActFrameSng < 1
            .ActFrameSng = .ActFrameSng + .NumFrames
        Loop
        
        Do While .ActFrameSng >= .NumFrames + 1
            .ActFrameSng = .ActFrameSng - .NumFrames
        Loop
                        
        .ActFrameInt = Int(.ActFrameSng)
    End With

End Sub

'clippt eine Animation und zeichnet diese
Public Sub Blit_Anim(ByVal X As Integer, ByVal Y As Integer, ByRef Surf As TSurf, ByRef Anim As TAnim)

    Dim h_Rect              As RECT
    Dim h_Right             As Integer
    Dim h_Bottom            As Integer

    With h_Rect
        .Left = Anim.FrameWidth * (Anim.ActFrameInt - 1)
        .Top = 0
        .Right = .Left + Anim.FrameWidth
        .Bottom = Surf.Height
        
        'rechter Rand
        h_Right = X + Anim.FrameWidth
        If h_Right > iXRes Then .Right = .Right - (h_Right - iXRes)
    
        'unterer Rand
        h_Bottom = Y + Surf.Height
        If h_Bottom > iYRes Then .Bottom = .Bottom - (h_Bottom - iYRes)

        'linker Rand
        If X < 0 Then
            .Left = .Left - X
            X = 0
        End If
    
        'oberer Rand
        If Y < 0 Then
            .Top = .Top - Y
            Y = 0
        End If
        
    End With
    
    'zeichnen
    g_BackBuf.BltFast X, Y, Surf.Surf, h_Rect, DDBLTFAST_DONOTWAIT Or DDBLTFAST_SRCCOLORKEY

End Sub

'ldt eine Bitmap-Schriftart
Public Function Load_Font(ByVal DescPath As String, ByVal SurfPath As String, ByRef Font As TFont, Optional ByVal KeyCol As Long = -1, Optional ByVal Width As Integer = -1, Optional ByVal Height As Integer = -1) As Boolean
    
    Dim n               As Long
    Dim FileNum         As Integer
    Dim h_X             As Long
    Dim h_Y             As Long
    Dim h_Width         As Long
    Dim h_Height        As Long
        
    On Error GoTo error:
    
    FileNum = FreeFile
        
    Open DescPath For Binary As #FileNum         'Description-Datei ffnen
        
        For n = 32 To 255
            
            With Font.Letter(n)
            
                Get #FileNum, , h_X
                Get #FileNum, , h_Y
                Get #FileNum, , h_Width
                Get #FileNum, , h_Height
                
                Font.Letter(n) = SetRectFrame(h_X, h_Y, h_Width, h_Height)
                
            End With
            
        Next
        
    Close #FileNum
    
    'Surface laden
    Load_Font = Load_Surf(SurfPath, Font.Surf, KeyCol, Width, Height)
    Exit Function
    
error:
    Load_Font = False

End Function

'entldt eine Schrift
Public Sub Unload_Font(ByRef Font As TFont)

    Unload_Surf Font.Surf

End Sub

'gibt Text aus (Bitmap-Font)
Public Sub Blit_Text(ByVal X As Integer, ByVal Y As Integer, ByVal Text As String, ByRef Font As TFont)

    Dim n                   As Long
    Dim h_Asc               As Integer
        
    With Font
        For n = 1 To Len(Text)
            h_Asc = Asc(Mid(Text, n, 1))
            
            If h_Asc >= 32 Then                                         'zeichnen
                Blit_SurfSector X, Y, .Surf, .Letter(h_Asc)
                
                'Position um einen Buchstaben verschieben
                X = X + .Letter(h_Asc).Right - .Letter(h_Asc).Left
            End If
        Next
    End With

End Sub

'Gibt die Pixel-Breite eines Textes zurck
Public Function GetTextWidth(ByVal Text As String, ByRef Font As TFont) As Integer
    
    Dim n               As Long
    Dim h_Asc           As Integer
    Dim h_Width         As Integer
    
    h_Width = 0
    
    For n = 1 To Len(Text)
        h_Asc = Asc(Mid(Text, n, 1))
            
        With Font.Letter(h_Asc)
            If h_Asc >= 32 Then h_Width = h_Width + .Right - .Left
        End With
    Next

    GetTextWidth = h_Width
    
End Function

'Gibt die Pixel-Hhe eines Textes zurck
Public Function GetTextHeight(ByRef Font As TFont) As Integer

    With Font.Letter(32)
        GetTextHeight = .Bottom - .Top
    End With

End Function

'Gibt einen TextBlock aus
Public Sub Blit_TextBlock(ByVal X As Integer, ByVal Y As Integer, ByVal Width As Integer, ByVal Text As String, ByRef Font As TFont, Optional ByVal Alignment As ETextAlignment = TEXT_ALIGNMENT_LEFT)

    Dim h_Pos(1 To 2)   As Long
    Dim m               As Long
    Dim h_WidthCnt      As Integer
    Dim h_HeightCnt     As Integer
    Dim h_TextWidth     As Integer
    Dim h_TextX         As Integer
    Dim h_OutputString  As String
    
    h_WidthCnt = 0
    h_HeightCnt = 0
    h_Pos(1) = 0
    h_Pos(2) = 0
    
    Do While h_Pos(2) < Len(Text)
        h_Pos(2) = h_Pos(2) + 1
        h_WidthCnt = h_WidthCnt + GetTextWidth(Mid(Text, h_Pos(2), 1), Font)
        
        If h_WidthCnt > Width Then
            
            'Breite zu gering fr einen Buchstaben
            If h_Pos(2) - h_Pos(1) = 1 Then Exit Sub
            
            'nach dem nchsten Leerzeichen suchen
            For m = h_Pos(2) To h_Pos(1) Step -1
                If m > h_Pos(1) Then
                    If Mid(Text, m, 1) = " " Then Exit For
                End If
            Next
        
            If m > h_Pos(1) Then
                h_Pos(2) = m
            Else
                h_Pos(2) = h_Pos(2) - 1
            End If
            
            'Text-Ausrichtung
            h_OutputString = Trim(Mid(Text, h_Pos(1) + 1, h_Pos(2) - h_Pos(1)))
            
            Select Case Alignment
                Case TEXT_ALIGNMENT_LEFT:
                    h_TextX = X
                Case TEXT_ALIGNMENT_RIGHT:
                    h_TextX = X + Width - GetTextWidth(h_OutputString, Font)
                Case TEXT_ALIGNMENT_CENTERED:
                    h_TextX = X + (Width - GetTextWidth(h_OutputString, Font)) * 0.5
            End Select
            
            'Ausgabe
            Blit_Text h_TextX, Y + h_HeightCnt, h_OutputString, Font
            
            h_WidthCnt = 0
            h_HeightCnt = h_HeightCnt + GetTextHeight(Font)
            h_Pos(1) = h_Pos(2)
        End If
    Loop
    
    'Text-Ausrichtung
    h_OutputString = Trim(Mid(Text, h_Pos(1) + 1, h_Pos(2) - h_Pos(1)))
    
    Select Case Alignment
        Case TEXT_ALIGNMENT_LEFT:
            h_TextX = X
        Case TEXT_ALIGNMENT_RIGHT:
            h_TextX = X + Width - GetTextWidth(h_OutputString, Font)
        Case TEXT_ALIGNMENT_CENTERED:
            h_TextX = X + (Width - GetTextWidth(h_OutputString, Font)) * 0.5
    End Select
    
    'Ausgabe
    Blit_Text h_TextX, Y + h_HeightCnt, h_OutputString, Font
    
End Sub

'fgt eine Msg zur Msg-Schlange dazu
Public Sub Add_Msg(ByVal X As Integer, ByVal Y As Integer, ByVal Text As String)
        
    MsgCount = MsgCount + 1
    
    ReDim Preserve MsgQueue(1 To MsgCount)
    
    With MsgQueue(MsgCount)
        .X = X
        .Y = Y
        .Text = Text
    End With
    
End Sub

'zeichnet alle Msgs
Public Sub Blit_Msgs()

    Dim n           As Long
    
    For n = 1 To MsgCount
        With MsgQueue(n)
            g_BackBuf.DrawText .X, .Y, .Text, False
        End With
    Next
    
    MsgCount = 0
    
End Sub

'lscht einen Teil des Bildschirms mit Farbe
Public Sub FillRect(ByRef Rectangle As RECT, ByVal Color As Long)
    
    g_BackBuf.BltColorFill Rectangle, Color
    
End Sub

'Ausgabe auf dem BildSchirm
Public Sub Flip(Optional ByVal VSync As Boolean = True)

    If VSync Then
        g_FrontBuf.Flip Nothing, DDFLIP_WAIT
    Else
        g_FrontBuf.Flip Nothing, DDFLIP_NOVSYNC
    End If
    
    CalcFPS
    
End Sub

'Ausgabe auf dem Bildschirm (Fenstermodus)
Public Sub FlipToWnd(ByRef DestRect As RECT, ByRef SrcRect As RECT)
        
    g_FrontBuf.Blt DestRect, g_BackBuf, SrcRect, DDBLT_WAIT
    
    CalcFPS

End Sub

'macht einen ScreenShot
Public Function ScreenShot(ByVal Path As String, ByRef PicBox As PictureBox) As Boolean

    Dim h_ScreenSurf        As DDSURFACEDESC2
    Dim DeskhWnd            As Long
    Dim DeskDC              As Long
      
    On Error GoTo error:
        
    With PicBox
        
        'Eigenschaften der Box setzen
        g_DD.GetDisplayMode h_ScreenSurf
        .Width = h_ScreenSurf.lWidth
        .Height = h_ScreenSurf.lHeight
        .BorderStyle = 0
        .Visible = False
        .AutoRedraw = True
        
        'Auf Bild zeichnen
        DeskhWnd = GetDesktopWindow()
        DeskDC = GetDC(DeskhWnd)
        BitBlt .hdc, 0, 0, .Width, .Height, DeskDC, 0, 0, vbSrcCopy
        ReleaseDC DeskhWnd, DeskDC
            
        'Bild speichern
        SavePicture .Image, Path
        .Cls
        
        ScreenShot = True
        Exit Function
    
    End With
    
error:
    ScreenShot = False
        
End Function

'aktiviert / deaktiviert den Cursor
Public Sub SetCursorVisibility(ByVal Visible As Boolean)
        
    If Visible Then
        Do Until ShowCursor(True) = 0
        Loop
    Else
        Do While ShowCursor(False) = -1
        Loop
    End If
        
End Sub

'wandelt RGB-Farbe in Long-Farbe um (nur fr 16 und 32 Bit)
Public Function CRGB(ByVal r As Byte, ByVal g As Byte, ByVal b As Byte, Optional ByVal ColDepth As Byte = 32, Optional ByVal Reverse As Boolean = True) As Long
    
    If Reverse Then
        
        Select Case ColDepth
                
            Case 32:
                CRGB = b + BitDisp(g, 8) + BitDisp(r, 16)
                
            Case 16:
                CRGB = BitDisp(b, -3) + BitDisp(BitDisp(g, -2), 5) + BitDisp(BitDisp(r, -3), 11)
            
            Case Else:
                CRGB = 0
        
        End Select
        
    Else
    
        Select Case ColDepth
                
            Case 32:
                CRGB = r + BitDisp(g, 8) + BitDisp(b, 16)
                
            Case 16:
                CRGB = BitDisp(r, -3) + BitDisp(BitDisp(g, -2), 5) + BitDisp(BitDisp(b, -3), 11)
            
            Case Else:
                CRGB = 0
        
        End Select
    
    End If
    
End Function

'Bitverschiebung (nur fr positive Zahlen)
Private Function BitDisp(ByVal Val As Long, ByVal Disp As Integer) As Long

    Dim Bit(0 To 30)    As Byte
    Dim LngVal          As Long
    Dim n               As Long
    Dim SgnDisp         As Integer
    
    'in BitZahl umrechnen
    For n = 30 To 0 Step -1
        If Val \ (2 ^ n) Then
            Bit(n) = 1
            Val = Val - 2 ^ n
        Else
            Bit(n) = 0
        End If
    Next
    
    'verschieben
    SgnDisp = Sgn(Disp)
    Disp = Abs(Disp)
    
    If SgnDisp = 1 Then         'nach links verschieben
        For n = 30 To Disp Step -1
            Bit(n) = Bit(n - Disp)
        Next
        
        For n = 0 To Disp - 1
            Bit(n) = 0
        Next
    ElseIf SgnDisp = -1 Then    'nach rechts verschieben
        For n = 0 To 30 - Disp
            Bit(n) = Bit(n + Disp)
        Next
        
        For n = 30 - Disp + 1 To 30
            Bit(n) = 0
        Next
    End If
    
    'zurckrechnen
    LngVal = 0
    
    For n = 0 To 30
        LngVal = LngVal + Bit(n) * (2 ^ n)
    Next
    
    BitDisp = LngVal

End Function

'gibt die aktuelle XAuflsung zurck
Public Function GetXRes() As Integer

    GetXRes = iXRes

End Function

'gibt die aktuelle YAuflsung zurck
Public Function GetYRes() As Integer

    GetYRes = iYRes

End Function

'gibt die aktuelle Farbtiefe zurck
Public Function GetColDepth() As Byte

    GetColDepth = iColDepth

End Function

'berechnet die FPS
Private Sub CalcFPS()
    
    Dim CurrentCntrVal      As Currency

    QueryPerformanceCounter CurrentCntrVal

    If LastCntrVal < CurrentCntrVal Then
        FramesPerSec = FPSLoopQuotient / ((CurrentCntrVal - LastCntrVal) / TimerFreq)
        LastCntrVal = CurrentCntrVal
        FPSLoopQuotient = 1
    Else
        FPSLoopQuotient = FPSLoopQuotient + 1
    End If
        
    If FramesPerSec < 1 Then FramesPerSec = 1
    AVSpeed = 100 / FramesPerSec
        
End Sub

'gibt FPS zurck
Public Function GetFPS() As Single
    
    GetFPS = FramesPerSec
        
End Function

'gibt den Durchschnitt eines Wertes (Abhngig von der FrameRate) zurck
Public Function AV(ByVal Val As Single) As Single

    AV = Val * AVSpeed

End Function

Public Sub CreateTexture(ByRef Path As String, ByRef Tex As DirectDrawSurface7)

    Dim ddsd3 As DDSURFACEDESC2
    Dim TextureEnum As Direct3DEnumPixelFormats
    
    Set TextureEnum = g_D3DDev.GetTextureFormatsEnum()
    TextureEnum.GetItem 1, ddsd3.ddpfPixelFormat
    ddsd3.ddsCaps.lCaps = DDSCAPS_TEXTURE
    Set Tex = g_DD.CreateSurfaceFromFile(Path, ddsd3)
    
End Sub

'Initialisiert Direct3D und gibt zurck ob Vorgang erfolgreich war
Public Function Init_D3D(ByVal ResX As Integer, ByVal ResY As Integer) As Boolean
    
    Dim n                   As Long
    Dim ViewPort            As D3DVIEWPORT7
    
    On Error GoTo error:
    
    Set g_D3D = g_DD.GetDirect3D
    Set g_D3DDev = g_D3D.CreateDevice("IID_IDirect3DHALDevice", g_BackBuf)
    
    '3D-Render-Bereich deklarieren
    With ViewPort
        .lX = 0
        .lY = 0
        .lWidth = ResX
        .lHeight = ResY
    End With
        
    
    g_D3DDev.SetViewport ViewPort
    
    With g_RectViewport(0)
        .X1 = 0
        .Y1 = 0
        .X2 = ResX
        .Y2 = ResY
    End With
    
    'BildschirmPosition setzen
    g_DX.IdentityMatrix g_matView
                              
    g_DX.ViewMatrix g_matView, SetVector(0, 0, -30), SetVector(0, 0, 0), SetVector(0, 1, 0), 0
    
    g_D3DDev.SetTransform D3DTRANSFORMSTATE_VIEW, g_matView
      
    'Projektion der 3D-Welt festlegen
    g_DX.IdentityMatrix g_matProj
        
    g_DX.ProjectionMatrix g_matProj, 1, 100, PI / 3
    g_D3DDev.SetTransform D3DTRANSFORMSTATE_PROJECTION, g_matProj
    
    g_D3DDev.SetRenderTarget g_BackBuf
    
    g_D3DDev.SetRenderState D3DRENDERSTATE_CULLMODE, D3DCULL_NONE
    
    Set DDI = g_DD.GetDeviceIdentifier(DDGDI_DEFAULT)
    
    D3DDivX = 29.5615 * iXRes / 1024
    D3DDivY = 22.171 * iYRes / 768
    
    D3DSubX = -((iXRes + 1) / D3DDivX) / 2
    D3DSubY = ((iYRes + 1) / D3DDivY) / 2
    
    Init_D3D = True
    Exit Function
    
error:
    Init_D3D = False

End Function

'fhrt Direct3D herunter
Public Sub Unload_D3D()

    Set g_D3DDev = Nothing
    Set g_D3D = Nothing

End Sub

Public Sub InitEffects()
    Dim n As Long

    g_DX.CreateD3DVertex 0, 0, 0, 0, 0, -1, 0, 0, FXVertex(1)
    g_DX.CreateD3DVertex 0, 0, 0, 0, 0, -1, 1, 0, FXVertex(2)
    g_DX.CreateD3DVertex 0, 0, 0, 0, 0, -1, 0, 1, FXVertex(3)
    g_DX.CreateD3DVertex 0, 0, 0, 0, 0, -1, 1, 1, FXVertex(4)
    
End Sub

Public Sub SetCam()

    'BildschirmPosition setzen
    g_DX.IdentityMatrix g_matView
                              
    g_DX.ViewMatrix g_matView, SetVector(0, 0, -30), SetVector(0, 0, 0), SetVector(0, 1, 0), 0
    
    g_D3DDev.SetTransform D3DTRANSFORMSTATE_VIEW, g_matView
      
    'Projektion der 3D-Welt festlegen
    g_DX.IdentityMatrix g_matProj
        
    g_DX.ProjectionMatrix g_matProj, 1, 100, PI / 3
    g_D3DDev.SetTransform D3DTRANSFORMSTATE_PROJECTION, g_matProj
        
End Sub

Public Function IsEffectInRect(iRect As RECT, ByRef iEffect As TEffect, X As Single, Y As Single) As Boolean
    Dim r As Single
    
    IsEffectInRect = False
    
    With iEffect
        r = .StartMaxRadius
        If r < .StartMinRadius Then r = .StartMinRadius
        If r < .EndRadius Then r = .EndRadius
        
        If IsInRectR(iRect, X, Y, r) Then
           
            IsEffectInRect = True
           
        End If
    End With
    
End Function

Private Sub PrepareEffectMaterial(iEffect As TEffect, ByRef T As Single)
    
    With iEffect
        Select Case .ColorMode
        Case ONLY_START_COLOR
                SetMaterialE .StartColor.r - T, .StartColor.g - T, .StartColor.b - T
        Case ALL_COLORS
            If T < 0.333 Then
                SetMaterialE .StartColor.r - T, .StartColor.g - T, .StartColor.b - T
            ElseIf T < 0.666 Then
                SetMaterialE T - .InvColor.b, T - .InvColor.r, T - .InvColor.g
            Else
                SetMaterialE .EndColor.r - T, .EndColor.g - T, .EndColor.b - T
            End If
        Case START_AND_INV_COLOR
            If T < 0.4 Then
                SetMaterialE T - .StartColor.b, T - .StartColor.r, T - .StartColor.g
            Else
                SetMaterialE .InvColor.r - T, .InvColor.g - T, .InvColor.b - T
            End If
        End Select
    End With
    
End Sub


Private Sub SetMaterialE(ByRef r As Single, ByRef g As Single, ByRef b As Single)

    Material.emissive.r = r
    Material.emissive.g = g
    Material.emissive.b = b

    g_D3DDev.SetMaterial Material

End Sub
    

Private Sub SetAlpha(ByRef Alpha As Long, ByRef Effect As EAlphaEffect, ByRef Textured As Boolean)
    Dim Src As Long
    Dim Dest As Long
    
    g_D3DDev.SetRenderState D3DRENDERSTATE_ALPHABLENDENABLE, Alpha
        
    If InStr(1, DDI.GetDescription, "Voodoo", vbBinaryCompare) > 0 Then
        Select Case Effect
        Case A_OFF
            Src = 1
            Dest = 1
        Case A_ADD
            Src = 2
            Dest = 2
        Case A_MULTIPLY
            Src = 3
            Dest = 5
        Case A_SUBTRACT
            Src = 3
            Dest = 7
        Case A_INVERT
            Src = 10
            Dest = 4
        End Select
    Else
        If Textured Then
            Select Case Effect
            Case A_OFF
                Src = 2
                Dest = 1
            Case A_ADD
                Src = 2
                Dest = 2
            Case A_MULTIPLY
                Src = 9
                Dest = 5
            Case A_SUBTRACT
                Src = 6
                Dest = 3
            Case A_INVERT
                Src = 10
                Dest = 4
            End Select
        Else
            Select Case Effect
            Case A_OFF
                Src = 2
                Dest = 1
            Case A_ADD
                Src = 2
                Dest = 2
            Case A_MULTIPLY
                Src = 9
                Dest = 6
            Case A_SUBTRACT
                Src = 9
                Dest = 5
            Case A_INVERT
                Src = 10
                Dest = 4
            End Select
        End If
    End If
    
    g_D3DDev.SetRenderState D3DRENDERSTATE_SRCBLEND, Src
    g_D3DDev.SetRenderState D3DRENDERSTATE_DESTBLEND, Dest

End Sub
Private Function IsInRectR(iRect As RECT, X As Single, Y As Single, r As Single) As Boolean

    IsInRectR = False

    If X + r > iRect.Left And X - r < iRect.Right And Y + r > iRect.Top And Y - r < iRect.Bottom Then
        IsInRectR = True
    End If

End Function


Public Function SetColor(r As Single, g As Single, b As Single) As D3DCOLORVALUE

    SetColor.r = r / 255
    SetColor.g = g / 255
    SetColor.b = b / 255

End Function

'definiert einen D3D-Vector
Private Function SetVector(ByVal a As Single, ByVal b As Single, ByVal c As Single) As D3DVECTOR
    
    With SetVector
        .X = a
        .Y = b
        .z = c
    End With
    
End Function


'Pythagoras (Addition)
Private Function Pyth(ByVal a As Single, ByVal b As Single) As Single

    Pyth = Sqr(a * a + b * b)

End Function

'#########################################################################################################
'#########################################################################################################
'#########################################################################################################
'ParticleEffects

Public Sub SetEffect(ByRef iEffect As TEffect, ByRef X As Single, ByRef Y As Single)
    Dim k As Long
    
    With iEffect
            
        .Draw = True
        
        .X = X
        .Y = Y
        
         If .InfiniteSpread Then
            
                .CurrentParticle = 0
                
                SetParticle iEffect.Particle(1), iEffect
            
        Else
        
            For k = 1 To .ParticleCount
            
                SetParticle .Particle(k), iEffect
                
            Next k
        
        End If
        
    End With

End Sub

Public Sub DrawEffect(ByRef iEffect As TEffect, ByRef Texture As DirectDrawSurface7)
    Dim n                   As Long
    Dim k                   As Long
    Dim RX                  As Single
    Dim RY                  As Single
    Dim CSlow               As Single
    
    CSlow = 1 - 0.01 * AVSpeed
        
    SetAlpha 1, A_ADD, True
    
    g_D3DDev.BeginScene
    
        g_D3DDev.SetTexture 0, Texture
    
        With iEffect
            
            If .InfiniteSpread Then
            
                If .CurAddParticle >= 1 Then .CurAddParticle = 0
                
                .CurAddParticle = .CurAddParticle + .ParticleCount * .AttenuationSpeed * AVSpeed
                
                If .CurAddParticle > 1 Then .CurAddParticle = Int(.CurAddParticle)
                
                If .CurAddParticle > .ParticleCount Then .CurAddParticle = .ParticleCount
            
                .CurrentParticle = .CurrentParticle + .CurAddParticle
                If .CurrentParticle > .ParticleCount - .CurAddParticle Then .CurrentParticle = 0
                
                For k = .CurrentParticle + 1 To .CurrentParticle + .CurAddParticle
                
                    SetParticle iEffect.Particle(k), iEffect
                                    
                Next k
            Else
            
                If .Particle(6).Attenuation > 1.1 Then
                    If .ParticleCount > 0 And .CurrentParticle = 0 Then
                        If .Particle(.ParticleCount - 6).Attenuation > 1.1 Then .Draw = False
                    End If
                End If

            End If
    
        End With
        
        If iEffect.Draw Then
            For n = 1 To iEffect.ParticleCount
                With iEffect.Particle(n)
            
                    If .Attenuation > iEffect.VDecelThreshold And iEffect.VDeceleration <> 1 Then
                        .MX = .MX * iEffect.VDeceleration * CSlow
                        .MY = .MY * iEffect.VDeceleration * CSlow
                    End If
                        
                    .MX = .MX + iEffect.AccX * AVSpeed
                    .MY = .MY + iEffect.AccY * AVSpeed
            
                    .VX = .VX + .MX * AVSpeed
                    .VY = .VY + .MY * AVSpeed
                    
                    .Rad = .Rad + .RS * AVSpeed
                    
                    If iEffect.AttenuationExpMode Then
                        .Attenuation = (.Attenuation ^ 2 + iEffect.AttenuationSpeed * AVSpeed) / .Attenuation
                    Else
                        .Attenuation = .Attenuation + iEffect.AttenuationSpeed * AVSpeed
                    End If
                    
                    If IsInRectR(ScreenRect, iEffect.X + .VX, iEffect.Y + .VY, iEffect.MaxParticleRadius * 15) Then
                                            
                        RX = (iEffect.X + .VX) / D3DDivX + D3DSubX
                        RY = -(iEffect.Y + .VY) / D3DDivY + D3DSubY
    
                        If .RS = 0 Then
                            FXVertex(1).X = RX - .Radius
                            FXVertex(1).Y = RY + .Radius
        
                            FXVertex(2).X = RX + .Radius
                            FXVertex(2).Y = RY + .Radius
        
                            FXVertex(3).X = RX - .Radius
                            FXVertex(3).Y = RY - .Radius
        
                            FXVertex(4).X = RX + .Radius
                            FXVertex(4).Y = RY - .Radius
                        Else
                            FXVertex(1).X = RX + Cos(.Rad) * .Radius
                            FXVertex(1).Y = RY + Sin(.Rad) * .Radius
        
                            FXVertex(2).X = RX + Cos(.Rad + HPI) * .Radius
                            FXVertex(2).Y = RY + Sin(.Rad + HPI) * .Radius
        
                            FXVertex(3).X = RX + Cos(.Rad + HPI + PI) * .Radius
                            FXVertex(3).Y = RY + Sin(.Rad + HPI + PI) * .Radius
        
                            FXVertex(4).X = RX + Cos(.Rad + PI) * .Radius
                            FXVertex(4).Y = RY + Sin(.Rad + PI) * .Radius
                        End If
                        
                        PrepareEffectMaterial iEffect, .Attenuation
                        
                        g_D3DDev.DrawPrimitive D3DPT_TRIANGLESTRIP, D3DFVF_LVERTEX, FXVertex(1), 4, D3DDP_DEFAULT
                                                    
                    End If
                                    
                End With
            Next n
        End If
           
    g_D3DDev.EndScene
    
End Sub

Private Sub SetParticle(ByRef Particle As TEffectParticle, ByRef iEffect As TEffect)
    Dim r As Single
    Dim Rad As Single
    Dim inc As Single
    Dim v As Single
    Dim a As Single
    
    With Particle
        Rad = Rnd * iEffect.SpreadRange + iEffect.SpreadRangeOffset
        inc = Rnd * (iEffect.VMaxIncreasing - iEffect.VMinIncreasing) + iEffect.VMinIncreasing
        
        v = ((iEffect.EndRadius - iEffect.StartMinRadius) * iEffect.AttenuationSpeed)
         
        If (iEffect.VMaxIncreasing = 0 And iEffect.VMinIncreasing = 0) Then inc = 1
        
        If iEffect.VDivIncreasing Then
            .MX = (Cos(Rad)) * v / inc + iEffect.VelX
            .MY = (Sin(Rad)) * v / inc + iEffect.VelY
        Else
            .MX = (Cos(Rad)) * v * inc + iEffect.VelX
            .MY = (Sin(Rad)) * v * inc + iEffect.VelY
        End If
        
        r = Rnd * (iEffect.StartMaxRadius - iEffect.StartMinRadius) + iEffect.StartMinRadius
        
        .VX = Cos(Rad) * r
        .VY = Sin(Rad) * r

        .RS = Rnd * (iEffect.MaxParticleRot - iEffect.MinParticleRot) + iEffect.MinParticleRot

        .Attenuation = Rnd * iEffect.AttenuationStartRange

        .Radius = Rnd * (iEffect.MaxParticleRadius - iEffect.MinParticleRadius) + iEffect.MinParticleRadius
        
    End With

End Sub

'#########################################################################################################
'#########################################################################################################
'#########################################################################################################
'Lights

Public Sub SetLight(ByRef iLight As TLight, ByRef Attenuation As Single, ByRef X As Single, ByRef Y As Single)
    
    With iLight
        
        .Draw = True
        
        .Attenuation = Attenuation
        
        .X = X
        .Y = Y
        
         .iRadius = (.Radius / D3DDivX)
        
    End With

End Sub

Public Sub DrawLight(ByRef iLight As TLight, ByRef Texture As DirectDrawSurface7)
    Dim n                   As Long
    Dim RX                  As Single
    Dim RY                  As Single
    Dim f                   As Single
    Dim CSlow               As Single
    
    CSlow = 1 - 0.01 * AVSpeed
    
    SetAlpha 1, A_MULTIPLY, True
    
    g_D3DDev.BeginScene
    
        g_D3DDev.SetTexture 0, Texture
    
        With iLight
                       
            If .Attenuation > 1 Then .Draw = False
              
            If .Draw Then
                
                If .Attenuation > .VDecelThreshold Then
                    .VelX = .VelX * .VDeceleration * CSlow
                    .VelY = .VelY * .VDeceleration * CSlow
                End If
                    
                .VelX = .VelX + .AccX * AVSpeed
                .VelY = .VelY + .AccY * AVSpeed
        
                .X = .X + .VelX * AVSpeed
                .Y = .Y + .VelY * AVSpeed
                
                .Attenuation = .Attenuation + .AttenuationSpeed * AVSpeed
                
                If IsInRectR(ScreenRect, .X, .Y, .Radius) Then
                                        
                    RX = .X / D3DDivX + D3DSubX
                    RY = -.Y / D3DDivY + D3DSubY
        
                    FXVertex(1).X = RX - .iRadius
                    FXVertex(1).Y = RY + .iRadius * 1.3
    
                    FXVertex(2).X = RX + .iRadius
                    FXVertex(2).Y = RY + .iRadius * 1.3
    
                    FXVertex(3).X = RX - .iRadius
                    FXVertex(3).Y = RY - .iRadius * 1.3
    
                    FXVertex(4).X = RX + .iRadius
                    FXVertex(4).Y = RY - .iRadius * 1.3
                
                    SetMaterialE .Color.r - .Attenuation, .Color.g - .Attenuation, .Color.b - .Attenuation
                    
                    For n = 1 To CLng(.Intensity)
                        g_D3DDev.DrawPrimitive D3DPT_TRIANGLESTRIP, D3DFVF_LVERTEX, FXVertex(1), 4, D3DDP_DEFAULT
                    Next
                    
                    f = .Intensity - CLng(.Intensity)
                    If f > 0 Then
                        SetMaterialE (.Color.r - .Attenuation) * f, (.Color.g - .Attenuation) * f, (.Color.b - .Attenuation) * f
                        g_D3DDev.DrawPrimitive D3DPT_TRIANGLESTRIP, D3DFVF_LVERTEX, FXVertex(1), 4, D3DDP_DEFAULT
                    End If
                    
                End If
                
            End If
                                                    
        End With
               
    g_D3DDev.EndScene
    
End Sub

'#########################################################################################################
'#########################################################################################################
'#########################################################################################################
'Snow

Public Sub InitSnow(ByRef MaxParticles As Long, ByRef Texture As String)

    Dim n As Long
        
    CreateTexture Texture, SnowTexture
    
    MAX_SNOW_PARTICLES = MaxParticles
    
    ReDim SnowParticle(1 To MAX_SNOW_PARTICLES)
    
    For n = 1 To MAX_SNOW_PARTICLES
    
        With SnowParticle(n)
        
            .VX = Rnd * iXRes
            .VY = Rnd * iYRes
            
            .Rad = Rnd
            .RS = 0.01 * Rnd - 0.005
            
            .MX = 0
            .MY = (.RS + 0.005) * 100
            
            .Radius = .MY / 4 + 0.1
    
        End With
    
    Next n

End Sub


Public Sub DrawSnow(ByRef Attenuation As Single, ByRef ScreenPosX As Long, ByRef ScreenPosY As Long)
    Dim n                   As Long
    Dim k                   As Long
    Dim RX                  As Single
    Dim RY                  As Single
    Dim PFactor             As Single
    
    PFactor = 1.5
    
    SetAlpha 1, A_ADD, True
    
    g_D3DDev.BeginScene
    
        g_D3DDev.SetTexture 0, SnowTexture
        
        For n = 1 To MAX_SNOW_PARTICLES
            With SnowParticle(n)
            
                If .VX + .Radius * D3DDivX < ScreenPosX * PFactor Then
                    .VX = .VX + iXRes + .Radius * D3DDivX * 2
                ElseIf .VX - 0.1 * D3DDivX > ScreenPosX * PFactor + iXRes Then
                    .VX = .VX - iXRes - .Radius * D3DDivX * 2
                End If
                
                If .VY + .Radius * D3DDivY < ScreenPosY * PFactor Then
                    .VY = .VY + iYRes + .Radius * D3DDivY * 2
                ElseIf .VY - .Radius * D3DDivY > ScreenPosY * PFactor + iYRes Then
                    .VY = .VY - iYRes - .Radius * D3DDivY * 2
                End If
            
                .MX = Sin(.Rad) * .Radius
                                
                .VX = .VX + .MX * AVSpeed
                .VY = .VY + .MY * AVSpeed
                
                .Rad = .Rad + .RS * AVSpeed
                                        
                 RX = (.VX - ScreenPosX * PFactor) / D3DDivX + D3DSubX
                 RY = -(.VY - ScreenPosY * PFactor) / D3DDivY + D3DSubY

                 FXVertex(1).X = RX + Cos(.Rad) * .Radius
                 FXVertex(1).Y = RY + Sin(.Rad) * .Radius

                 FXVertex(2).X = RX + Cos(.Rad + HPI) * .Radius
                 FXVertex(2).Y = RY + Sin(.Rad + HPI) * .Radius

                 FXVertex(3).X = RX + Cos(.Rad + HPI + PI) * .Radius
                 FXVertex(3).Y = RY + Sin(.Rad + HPI + PI) * .Radius

                 FXVertex(4).X = RX + Cos(.Rad + PI) * .Radius
                 FXVertex(4).Y = RY + Sin(.Rad + PI) * .Radius
                 
                 SetMaterialE Attenuation, Attenuation, Attenuation
            
                 g_D3DDev.DrawPrimitive D3DPT_TRIANGLESTRIP, D3DFVF_LVERTEX, FXVertex(1), 4, D3DDP_DEFAULT
                                
            End With
        Next n
           
    g_D3DDev.EndScene

End Sub

'#########################################################################################################
'#########################################################################################################
'#########################################################################################################
'ShockWaves

Public Sub InitWaveNet(ByRef Tex256x256 As String)
    Dim n As Long
    Dim m As Long
    Dim ddsd3 As DDSURFACEDESC2
    Dim TextureEnum As Direct3DEnumPixelFormats
    Dim dx As Long
    Dim dy As Long
    Dim tu1 As Single
    Dim tu2 As Single
    Dim tv1 As Single
    Dim tv2 As Single
   
    HNET_ELEMENTS = iXRes / ELEMENT_WIDTH
    VNET_ELEMENTS = iYRes / ELEMENT_HEIGHT

    HT_ELEMENTS = -Int(-iXRes / TEXTURE_WIDTH)
    VT_ELEMENTS = -Int(-iYRes / TEXTURE_HEIGHT)
    
    ReDim WaveNetV(1 To HNET_ELEMENTS, 1 To VNET_ELEMENTS)
    ReDim WaveNetT(1 To HT_ELEMENTS, 1 To VT_ELEMENTS)

    ScreenRadius = Pyth(HNET_ELEMENTS / 2, VNET_ELEMENTS / 2)
 
    'Texturen erstellen (laden)
    Set TextureEnum = g_D3DDev.GetTextureFormatsEnum()
    TextureEnum.GetItem 1, ddsd3.ddpfPixelFormat
    ddsd3.ddsCaps.lCaps = DDSCAPS_TEXTURE

    For n = 1 To HT_ELEMENTS
       For m = 1 To VT_ELEMENTS
    
            Set WaveNetT(n, m) = g_DD.CreateSurfaceFromFile(Tex256x256, ddsd3)
    
        Next m
    Next n
    
    'Texturkoordinaten einstellen
    dx = TEXTURE_WIDTH / ELEMENT_WIDTH
    dy = TEXTURE_HEIGHT / ELEMENT_HEIGHT
    
    For n = 0 To HNET_ELEMENTS - 1
        For m = 0 To VNET_ELEMENTS - 1
            
            tu1 = (n Mod dx) / dx
            tv1 = (m Mod dy) / dy
            tu2 = (n Mod dx + 1) / dx
            tv2 = (m Mod dy + 1) / dy
    
            With WaveNetV(n + 1, m + 1)
                g_DX.CreateD3DVertex 0, 0, 0, 0, 0, -1, tu1, tv1, .Vertex(1)
                g_DX.CreateD3DVertex 0, 0, 0, 0, 0, -1, tu2, tv1, .Vertex(2)
                g_DX.CreateD3DVertex 0, 0, 0, 0, 0, -1, tu1, tv2, .Vertex(3)
                g_DX.CreateD3DVertex 0, 0, 0, 0, 0, -1, tu2, tv2, .Vertex(4)
            End With
            
        Next m
        
    Next n
    
    'Vertices als Netz aufbauen
    Dim hf As Single
    Dim vf As Single
    
    Dim realHNET_ELEMENTS As Single
    Dim realVNET_ELEMENTS As Single
    
    hf = ELEMENT_WIDTH / D3DDivX
    vf = ELEMENT_HEIGHT / D3DDivY
    
    realHNET_ELEMENTS = iXRes / ELEMENT_WIDTH
    realVNET_ELEMENTS = iYRes / ELEMENT_HEIGHT
    
    For n = 1 To HNET_ELEMENTS
        For m = 1 To VNET_ELEMENTS
            With WaveNetV(n, m)
            
                .Vertex(1).X = (n - 1 - realHNET_ELEMENTS / 2) * hf
                .Vertex(1).Y = (-m + 1 + realVNET_ELEMENTS / 2) * vf
    
                .Vertex(2).X = (n - realHNET_ELEMENTS / 2) * hf
                .Vertex(2).Y = (-m + 1 + realVNET_ELEMENTS / 2) * vf
    
                .Vertex(3).X = (n - 1 - realHNET_ELEMENTS / 2) * hf
                .Vertex(3).Y = (-m + realVNET_ELEMENTS / 2) * vf
    
                .Vertex(4).X = (n - realHNET_ELEMENTS / 2) * hf
                .Vertex(4).Y = (-m + realVNET_ELEMENTS / 2) * vf
                                
            End With
        Next m
    Next n
    
End Sub

Public Sub SetWaveNet(ByRef iWave As T3DWave, ByRef X As Single, ByRef Y As Single)
    
    With iWave
        .Draw = True
        
        .MaxRad = .WaveCount * TPI
        .Rad = 0
        
        .X = X
        .Y = Y
        
    End With
    
End Sub

Public Sub DrawWaveNet()
    Dim n As Long
    Dim m As Long
    Dim Material As D3DMATERIAL7
    Dim dx As Long
    Dim dy As Long
    
    g_D3DDev.SetTextureStageState 0, D3DTSS_MAGFILTER, 1
    g_D3DDev.SetTextureStageState 0, D3DTSS_MINFILTER, 1
    
    RefreshWaveTextures
        
    dx = TEXTURE_WIDTH / ELEMENT_WIDTH
    dy = TEXTURE_HEIGHT / ELEMENT_HEIGHT
    
    g_D3DDev.BeginScene
        
        SetAlpha 0, A_OFF, True
    
        SetMaterialE 1, 1, 1
        
        For n = 1 To HNET_ELEMENTS
            For m = 1 To VNET_ELEMENTS
            
                If WaveNetV(n, m).Draw Then
                
                    g_D3DDev.SetTexture 0, WaveNetT((n - 1) \ dx + 1, (m - 1) \ dy + 1)
                
                    g_D3DDev.DrawPrimitive D3DPT_TRIANGLESTRIP, D3DFVF_VERTEX, WaveNetV(n, m).Vertex(1), 4, D3DDP_DEFAULT
                    
                End If
                    
            Next m
        Next n
        
    g_D3DDev.EndScene

End Sub

Public Sub MoveWaves(ByRef iWave() As T3DWave, ByRef iBegin As Long, ByRef iEnd As Long)
    Dim n As Long
    Dim m As Long
    Dim k As Long
    
    Dim MinOneWave As Boolean
    
    Dim H As Single
    Dim d As Single
    
    Dim dx As Long
    Dim dy As Long
    
    MinOneWave = False
    
    'Verarbeitung der WaveProperties
    For k = iBegin To iEnd
        With iWave(k)

            If .c > 0 Then
                .c = .c - .cAttenuation * AVSpeed
            Else
                .c = 0
            End If
            .Rad = .Rad - .c * AVSpeed
        
            If .Amplitude > 0 Then
                .Amplitude = .Amplitude - .AmpAttenuation * AVSpeed
            Else
                .Amplitude = 0
                .Draw = False
            End If
        
            .SrcX = .X \ ELEMENT_WIDTH
            .SrcY = .Y \ ELEMENT_HEIGHT
            
            dx = (iXRes / 2) / ELEMENT_WIDTH
            dy = (iYRes / 2) / ELEMENT_HEIGHT
            
            If RadiusCollision(.X \ ELEMENT_WIDTH, .Y \ ELEMENT_HEIGHT, -(.Rad), dx, dy, ScreenRadius) And _
               Not CircleInCircle(.X \ ELEMENT_WIDTH, .Y \ ELEMENT_HEIGHT, -(.Rad + .WaveCount * .WaveLength * TPI), dx, dy, ScreenRadius) Then
                .Draw = True
            Else
                .Draw = False
            End If
            
            If .Draw Then MinOneWave = True

        End With
    Next k
    
    If MinOneWave Then              'wenn mindestens eine gezeichnet werden muss
    
        'Hhen werden errechnet
        For n = 1 To HNET_ELEMENTS
            For m = 1 To VNET_ELEMENTS
                    
                WaveNetV(n, m).Height = 0
                
                For k = iBegin To iEnd
                    With iWave(k)
                        If .Draw Then
                    
                            H = .Amplitude
                            d = (Pyth(.SrcX - n, .SrcY - m) + .Rad) / .WaveLength
                            
                            If d > 0 Then H = 0
                            
                            If d < -.MaxRad Then H = 0
                            
                            If H Then
                            
                                WaveNetV(n, m).Height = WaveNetV(n, m).Height + Cos(d) * H - H
                                    
                            End If
                            
                        End If
                    End With
                Next k
                    
            Next m
        Next n
        
        'Vertices werden Angeglichen
        Dim n1 As Long
        Dim m1 As Long
        Dim tmph As Single
        For n = 1 To HNET_ELEMENTS
            For m = 1 To VNET_ELEMENTS
                With WaveNetV(n, m)
                    
                        n1 = n + 1
                        m1 = m + 1
                        
                        If n1 > HNET_ELEMENTS Then n1 = n
                        If m1 > VNET_ELEMENTS Then m1 = m
                    
                        .Vertex(1).z = .Height
                        .Vertex(2).z = WaveNetV(n1, m).Height
                        .Vertex(3).z = WaveNetV(n, m1).Height
                        .Vertex(4).z = WaveNetV(n1, m1).Height
                        
                            
                        tmph = 0
                        tmph = .Vertex(1).z + .Vertex(2).z + .Vertex(3).z + .Vertex(4).z
                        
                        If tmph Then
                            .Draw = True
                        Else
                            .Draw = False
                        End If
                        
                
                End With
            Next m
        Next n

    End If
    
End Sub

Private Sub RefreshWaveTextures()
    Dim n As Long
    Dim m As Long
    Dim tmpRect As RECT
    
    For n = 0 To HT_ELEMENTS - 1
        For m = 0 To VT_ELEMENTS - 1
            tmpRect.Left = n * TEXTURE_WIDTH
            tmpRect.Top = m * TEXTURE_HEIGHT
            tmpRect.Right = (n + 1) * TEXTURE_WIDTH
            tmpRect.Bottom = (m + 1) * TEXTURE_HEIGHT
            
            If tmpRect.Right > iXRes Then tmpRect.Right = iXRes
            If tmpRect.Bottom > iYRes Then tmpRect.Bottom = iYRes
        
            WaveNetT(n + 1, m + 1).BltFast 0, 0, g_BackBuf, tmpRect, DDBLTFAST_WAIT
        Next m
    Next n
    
End Sub

Public Function AreWavesVisible(ByRef iWave() As T3DWave, ByRef iBegin As Long, ByRef iEnd As Long) As Boolean
    Dim k As Long
    
    AreWavesVisible = False
    
    For k = iBegin To iEnd
        With iWave(k)
            
            If .Draw Then
                AreWavesVisible = True
                Exit For
            End If
            
        End With
    Next k

End Function


'Radius Kollision
Private Function RadiusCollision(ByRef MidX1 As Long, ByRef MidY1 As Long, ByRef r1 As Single, ByRef MidX2 As Long, ByRef MidY2 As Long, ByRef r2 As Single) As Boolean
    
    Dim a As Long
    Dim b As Long
    
    a = MidX2 - MidX1
    b = MidY2 - MidY1
    RadiusCollision = (Sqr(a * a + b * b) <= r1 + r2)
        
End Function

'Test ob Kreis in Kreis ist
Private Function CircleInCircle(ByRef MidX1 As Long, ByRef MidY1 As Long, ByRef r1 As Single, ByRef MidX2 As Long, ByRef MidY2 As Long, ByRef r2 As Single) As Boolean
    
    Dim a As Long
    Dim b As Long
    
    a = MidX2 - MidX1
    b = MidY2 - MidY1
    
    
    CircleInCircle = (r2 <= r1)
    
    If CircleInCircle Then CircleInCircle = (Sqr(a * a + b * b) + r2 < r1)
    
        
End Function

'#########################################################################################################
'#########################################################################################################
'#########################################################################################################
'MotionBlur

Public Sub InitBlurNet(ByRef Tex256x256 As String)
    Dim n As Long
    Dim m As Long
    Dim ddsd3 As DDSURFACEDESC2
    Dim TextureEnum As Direct3DEnumPixelFormats
    Dim dx As Long
    Dim dy As Long
    Dim tu1 As Single
    Dim tu2 As Single
    Dim tv1 As Single
    Dim tv2 As Single
           
    HNET_ELEMENTS = iXRes / ELEMENT_WIDTH
    VNET_ELEMENTS = iYRes / ELEMENT_HEIGHT

    HT_ELEMENTS = -Int(-iXRes / TEXTURE_WIDTH)
    VT_ELEMENTS = -Int(-iYRes / TEXTURE_HEIGHT)
 
    ReDim BlurNetV(1 To HNET_ELEMENTS, 1 To VNET_ELEMENTS)
    ReDim BlurNetT(1 To HT_ELEMENTS, 1 To VT_ELEMENTS)

    'Texturen erstellen (laden)
    Set TextureEnum = g_D3DDev.GetTextureFormatsEnum()
    TextureEnum.GetItem 1, ddsd3.ddpfPixelFormat
    ddsd3.ddsCaps.lCaps = DDSCAPS_TEXTURE

    For n = 1 To HT_ELEMENTS
       For m = 1 To VT_ELEMENTS
    
            Set BlurNetT(n, m) = g_DD.CreateSurfaceFromFile(Tex256x256, ddsd3)
    
        Next m
    Next n
    
    'Texturkoordinaten einstellen
    dx = TEXTURE_WIDTH / ELEMENT_WIDTH
    dy = TEXTURE_HEIGHT / ELEMENT_HEIGHT
    
    For n = 0 To HNET_ELEMENTS - 1
        For m = 0 To VNET_ELEMENTS - 1
            
            tu1 = (n Mod dx) / dx
            tv1 = (m Mod dy) / dy
            tu2 = (n Mod dx + 1) / dx
            tv2 = (m Mod dy + 1) / dy
    
            With BlurNetV(n + 1, m + 1)
                g_DX.CreateD3DVertex 0, 0, 0, 0, 0, -1, tu1, tv1, .Vertex(1)
                g_DX.CreateD3DVertex 0, 0, 0, 0, 0, -1, tu2, tv1, .Vertex(2)
                g_DX.CreateD3DVertex 0, 0, 0, 0, 0, -1, tu1, tv2, .Vertex(3)
                g_DX.CreateD3DVertex 0, 0, 0, 0, 0, -1, tu2, tv2, .Vertex(4)
            End With
            
        Next m
        
    Next n
    
    'Vertices als Netz aufbauen
    Dim hf As Single
    Dim vf As Single
    
    Dim realHNET_ELEMENTS As Single
    Dim realVNET_ELEMENTS As Single
    
    hf = ELEMENT_WIDTH / D3DDivX
    vf = (ELEMENT_HEIGHT) / D3DDivY
    
    realHNET_ELEMENTS = iXRes / ELEMENT_WIDTH
    realVNET_ELEMENTS = iYRes / ELEMENT_HEIGHT
    
    For n = 1 To HNET_ELEMENTS
        For m = 1 To VNET_ELEMENTS
            With BlurNetV(n, m)
            
                .Vertex(1).X = (n - 1 - realHNET_ELEMENTS / 2) * hf
                .Vertex(1).Y = (-m + 1 + realVNET_ELEMENTS / 2) * vf
    
                .Vertex(2).X = (n - realHNET_ELEMENTS / 2) * hf
                .Vertex(2).Y = (-m + 1 + realVNET_ELEMENTS / 2) * vf
    
                .Vertex(3).X = (n - 1 - realHNET_ELEMENTS / 2) * hf
                .Vertex(3).Y = (-m + realVNET_ELEMENTS / 2) * vf
    
                .Vertex(4).X = (n - realHNET_ELEMENTS / 2) * hf
                .Vertex(4).Y = (-m + realVNET_ELEMENTS / 2) * vf
                                
            End With
        Next m
    Next n
    
End Sub


Public Sub DrawBlurNet(ByRef BlurFactor As Single)
    Dim n As Long
    Dim m As Long
    Dim Material As D3DMATERIAL7
    Dim dx As Long
    Dim dy As Long
    
    g_D3DDev.SetTextureStageState 0, D3DTSS_MAGFILTER, 1
    g_D3DDev.SetTextureStageState 0, D3DTSS_MINFILTER, 1
        
    dx = TEXTURE_WIDTH / ELEMENT_WIDTH
    dy = TEXTURE_HEIGHT / ELEMENT_HEIGHT
    
    g_D3DDev.BeginScene
        
        SetAlpha 1, A_ADD, True

        SetMaterialE BlurFactor, BlurFactor, BlurFactor
        
        For n = 1 To HNET_ELEMENTS
            For m = 1 To VNET_ELEMENTS
                
                    g_D3DDev.SetTexture 0, BlurNetT((n - 1) \ dx + 1, (m - 1) \ dy + 1)
                
                    g_D3DDev.DrawPrimitive D3DPT_TRIANGLESTRIP, D3DFVF_VERTEX, BlurNetV(n, m).Vertex(1), 4, D3DDP_DEFAULT
                    
            Next m
        Next n
        
    g_D3DDev.EndScene

End Sub

Public Sub RefreshBlurTextures()
    Dim n As Long
    Dim m As Long
    Dim tmpRect As RECT
    
    For n = 0 To HT_ELEMENTS - 1
        For m = 0 To VT_ELEMENTS - 1
            tmpRect.Left = n * TEXTURE_WIDTH
            tmpRect.Top = m * TEXTURE_HEIGHT
            tmpRect.Right = (n + 1) * TEXTURE_WIDTH
            tmpRect.Bottom = (m + 1) * TEXTURE_HEIGHT
            
            If tmpRect.Right > iXRes Then tmpRect.Right = iXRes
            If tmpRect.Bottom > iYRes Then tmpRect.Bottom = iYRes
        
            BlurNetT(n + 1, m + 1).BltFast 0, 0, g_BackBuf, tmpRect, DDBLTFAST_WAIT
            
        Next m
    Next n
    
End Sub
