Attribute VB_Name = "GraFX_003_SOUND"
Option Explicit

'===Typen===

Public Type TSBuf                                               'Sound
    Buf                                 As DirectSoundBuffer
End Type


'===ffentliche Variablen===

Public g_DS                             As DirectSound          'DirectSound-Objekt


'===interne Variablen===

Private SBufDesc                        As DSBUFFERDESC
Private SWaveFormat                     As WAVEFORMATEX
Private SoundDistFactor                 As Single               'wie stark die Lautstrke eines Sounds bei Entfernung abnimmt

'initialisiert die Sound-Engine
Public Function Init_GraFX_SOUND(ByRef Frm As Form, Optional ByVal SoundDistFact = 2) As Boolean

    On Error GoTo error:
    
    SetSoundDistFactor SoundDistFact
    
    Set g_DS = g_DX.DirectSoundCreate("")
    g_DS.SetCooperativeLevel Frm.hWnd, DSSCL_PRIORITY
    
    With SWaveFormat
        .nFormatTag = WAVE_FORMAT_PCM
        .nChannels = 10
        .lSamplesPerSec = 22050
        .nBitsPerSample = 16
        .nBlockAlign = .nBitsPerSample / 8 * .nChannels
        .lAvgBytesPerSec = .lSamplesPerSec * .nBlockAlign
    End With
    
    'Einstellungsmglichkeiten
    SBufDesc.lFlags = DSBCAPS_CTRLFREQUENCY Or DSBCAPS_CTRLPAN Or DSBCAPS_CTRLVOLUME Or DSBCAPS_STATIC
    
    Init_GraFX_SOUND = True
    Exit Function
    
error:
    Init_GraFX_SOUND = False

End Function

'fhrt die Sound-Engine herunter
Public Sub Unload_GraFX_SOUND(ByRef Frm As Form)

    g_DS.SetCooperativeLevel Frm.hWnd, DSSCL_NORMAL
    
    Set g_DS = Nothing

End Sub

'ldt einen Soundbuffer in den Speicher
Public Function Load_SBuf(ByVal Path As String, ByRef SBuf As TSBuf) As Boolean

    On Error GoTo error:
    
    Set SBuf.Buf = g_DS.CreateSoundBufferFromFile(Path, SBufDesc, SWaveFormat)
    
    With SBuf.Buf
        .SetVolume 0
        .SetPan 0
        .SetCurrentPosition 0
    End With
    
    Load_SBuf = True
    Exit Function
    
error:
    Load_SBuf = False

End Function

'dupliziert einen Soundbuffer
Public Function Duplicate_SBuf(ByRef SBuf As TSBuf, ByRef OriginalSBuf As TSBuf) As Boolean

    On Error GoTo error:
    
    Set SBuf.Buf = g_DS.DuplicateSoundBuffer(OriginalSBuf.Buf)
    
    Duplicate_SBuf = True
    Exit Function
    
error:
    Duplicate_SBuf = False

End Function

'lscht einen SoundBuffer aus dem Speicher
Public Sub Unload_SBuf(ByRef SBuf As TSBuf)

    Set SBuf.Buf = Nothing

End Sub

'Setzt die Eigenschaften eines SoundBuffers
Public Sub Init_SBuf(ByRef SBuf As TSBuf, Optional ByVal Volume As Long = 1, Optional ByVal Pan As Long = 10001, Optional ByVal SpeedFact As Single = 1)
    
    Dim h_WaveFormatEx       As WAVEFORMATEX
    
    With SBuf.Buf
        If Volume <> 1 Then
            If Volume < -10000 Then Volume = -10000
            If Volume > 10000 Then Volume = 10000
            .SetVolume Volume
        End If
        
        If Pan <> 10001 Then
            If Pan < -10000 Then Pan = -10000
            If Pan > 10000 Then Pan = 10000
            .SetPan Pan
        End If
        
        'Frequenz errechnen
        .GetFormat h_WaveFormatEx
        .SetFrequency h_WaveFormatEx.lSamplesPerSec * SpeedFact
    End With
    
End Sub

'Setzt die Eigenschaften fr einen Stereo-Sound
Public Sub Init_Stereo_SBuf(ByRef SBuf As TSBuf, ByVal SrcX As Long, ByVal SrcY As Long, ByVal ListenerX As Long, ByVal ListenerY As Long, Optional ByVal SpeedFact As Single = 1)
        
    Dim h_Dist              As Long
    Dim h_Volume            As Long
    Dim h_Pan               As Long
    Dim h_WaveFormatEx      As WAVEFORMATEX
    Dim h_XDist             As Long
    Dim h_YDist             As Long
    
    With SBuf.Buf
        'Lautstrke errechnen
        h_XDist = ListenerX - SrcX
        h_YDist = ListenerY - SrcY
        h_Dist = Sqr(h_XDist * h_XDist + h_YDist * h_YDist)
        h_Volume = -(h_Dist * SoundDistFactor)
        
        If h_Volume < -10000 Then h_Volume = -10000
        .SetVolume h_Volume
        
        'Pan errechnen
        h_Pan = SrcX - ListenerX
        
        If h_Pan < -10000 Then
            h_Pan = -10000
        ElseIf h_Pan > 10000 Then
            h_Pan = 10000
        End If
                    
        .SetPan h_Pan
        
        'Frequenz errechnen
        .GetFormat h_WaveFormatEx
        .SetFrequency h_WaveFormatEx.lSamplesPerSec * SpeedFact
    End With

End Sub

'Spielt einen Sound ab
Public Sub Play_SBuf(ByRef SBuf As TSBuf, Optional ByVal LoopSBuf As Boolean = False)

    With SBuf.Buf
        If (.GetStatus And (DSBSTATUS_PLAYING Or _
        DSBSTATUS_LOOPING)) = 0 Then
        
            If LoopSBuf Then           'ob der Sound geloopt werden soll
                .Play DSBPLAY_LOOPING
            Else
                .Play DSBPLAY_DEFAULT
            End If
        
        End If
    End With
    
End Sub

'Stoppt einen laufenden Sound
Public Sub Stop_SBuf(ByRef SBuf As TSBuf)
    
    On Error Resume Next
    
    With SBuf.Buf
        If (.GetStatus And DSBSTATUS_PLAYING) Then .Stop
    End With
        
End Sub

'setzt PlayPos auf Anfang
Public Sub Reset_SBuf(ByRef SBuf As TSBuf)
    
    SBuf.Buf.SetCurrentPosition 0
    
End Sub

'setzt den SoundDistanzFaktor
Public Sub SetSoundDistFactor(ByVal Val As Single)
    
    SoundDistFactor = Val
    If SoundDistFactor < 0 Then SoundDistFactor = 0
    
End Sub

'gibt den SoundDistanzFaktor zurck
Public Function GetSoundDistFactor() As Single
    
    GetSoundDistFactor = SoundDistFactor
    
End Function
